from pathlib import Path
import tempfile
from math import isclose

from unittest import TestCase
from unittest.mock import patch

from obspy.core import Stream, read as read_ms
import numpy as np

from sohstationviewer.model.handling_data import (
    readSOHMSeed,
    readSOHTrace,
    readMPTrace,
    readWaveformTrace,
    readWaveformMSeed,
)
from sohstationviewer.model.reftek.from_rt2ms.core import Reftek130

TEST_DATA_DIR = Path(__file__).parent.parent.joinpath('test_data')


# tempfile.tempdir = './tempdir'


class TestHandlingData(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        rt130_dir = TEST_DATA_DIR.joinpath(
            'RT130-sample/2017149.92EB/2017150/92EB/9')
        cls.rt130_file = rt130_dir.joinpath('054910000_013EE8A0')
        rt130 = Reftek130.from_file(cls.rt130_file)
        cls.rt130_stream = Reftek130.to_stream(rt130)
        cls.rt130_trace = cls.rt130_stream[0]

        q330_dir = TEST_DATA_DIR.joinpath('Q330-sample/day_vols_AX08')
        cls.q330_soh_file = q330_dir.joinpath('AX08.XA..VKI.2021.186')
        cls.mseed_soh_stream = read_ms(cls.q330_soh_file)
        cls.mseed_soh_trace = cls.mseed_soh_stream[0]
        cls.q330_waveform_file = q330_dir.joinpath('AX08.XA..LHE.2021.186')
        cls.mseed_waveform_stream = read_ms(cls.q330_waveform_file)
        cls.mseed_waveform_trace = cls.mseed_waveform_stream[0]

    # @expectedFailure
    # def test_read_sohmseed(self):
    #     self.fail()

    def test_read_soh_trace_processed_trace_have_all_needed_info(self):
        processed_trace = readSOHTrace(self.mseed_soh_trace)
        with self.subTest('test_processed_trace_have_all_needed_info'):
            expected_key_list = [
                'chanID',
                'samplerate',
                'startTmEpoch',
                'endTmEpoch',
                'times',
                'data'
            ]
            self.assertTrue(
                all(key in processed_trace for key in expected_key_list),
            )

    def test_read_soh_trace_times_calculated_correctly(self):
        processed_trace = readSOHTrace(self.mseed_soh_trace)
        if isclose(processed_trace['startTmEpoch'], 0, abs_tol=0.0001):
            self.assertAlmostEqual(processed_trace['times'][0], 0)
        else:
            self.assertNotAlmostEqual(processed_trace['times'][0], 0)

    @patch('sohstationviewer.model.handling_data.readSOHTrace')
    def test_read_mp_trace(self, mock_read_soh_trace):
        mock_read_soh_trace.return_value = {
            # Contain five cases:
            # + Small positive
            # + Large positive
            # + Small negative
            # + Large negative
            # + Zero
            'data': np.array([1, 27272, -2, -23526, 0])
        }
        expected = np.array([0, 8.3, 0, -7.2, 0])
        processed_trace = readMPTrace(self.rt130_trace)
        self.assertTrue(
            np.array_equal(processed_trace['data'], expected)
        )

    @patch('sohstationviewer.model.handling_data.saveData2File')
    def test_read_waveform_trace(self, mock_save_data_2_file):
        station_id = self.rt130_trace.stats['station']
        channel_id = self.rt130_trace.stats['channel']
        # The function itself only cares about the length of this list so we
        # stub it out.
        traces_info = [dict() for _ in range(4)]
        tmp_dir = tempfile.TemporaryDirectory()
        processed_trace = readWaveformTrace(
            self.rt130_trace, station_id, channel_id, traces_info, tmp_dir.name
        )

        expected_key_list = [
            'samplerate',
            'startTmEpoch',
            'endTmEpoch',
            'size',
            'times_f',
            'data_f',
        ]
        self.assertTrue(
            all(key in processed_trace for key in expected_key_list),
        )

        self.assertTrue(mock_save_data_2_file.called)

    # @skip
    # def test_read_waveform_mseed(self):
    #     self.fail()
    #
    # @skip
    # def test_read_waveform_reftek(self):
    #     self.fail()
    #
    # @skip
    # def test_read_ascii(self):
    #     self.fail()
    #
    # @skip
    # def test_read_text(self):
    #     self.fail()
    #
    # @skip
    # def test_save_data2file(self):
    #     self.fail()
    #
    # @skip
    # def test_check_chan(self):
    #     self.fail()
    #
    # @skip
    # def test_check_sohchan(self):
    #     self.fail()
    #
    # @skip
    # def test_check_wfchan(self):
    #     self.fail()
    #
    # @skip
    # def test_sort_data(self):
    #     self.fail()
    #
    # @skip
    # def test_squash_gaps(self):
    #     self.fail()
    #
    # @skip
    # def test_downsample(self):
    #     self.fail()
    #
    # @skip
    # def test_constant_rate(self):
    #     self.fail()
    #
    # @skip
    # def test_chunk_minmax(self):
    #     self.fail()
    #
    # @skip
    # def test_trim_downsample_sohchan(self):
    #     self.fail()
    #
    # @skip
    # def test_trim_downsample_wfchan(self):
    #     self.fail()
    #
    # @skip
    # def test_get_each_day5min_list(self):
    #     self.fail()
    #
    # @skip
    # def test_get_trim_tpsdata(self):
    #     self.fail()
    #
    # @skip
    # def test_find_tpstm(self):
    #     self.fail()


class TestReadWaveformMSeed(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        q330_dir = TEST_DATA_DIR.joinpath('Q330-sample/day_vols_AX08')
        cls.q330_waveform_file = q330_dir.joinpath('AX08.XA..LHE.2021.186')
        cls.mseed_waveform_stream = read_ms(cls.q330_waveform_file)
        cls.mseed_waveform_trace = cls.mseed_waveform_stream[0]

    def setUp(self) -> None:
        self.station_id = self.mseed_waveform_trace.stats['station']
        self.channel_id = self.mseed_waveform_trace.stats['channel']
        # This list is only ever written to, so we can keep it empty
        self.traces_info = []

        self.temp_dir = tempfile.TemporaryDirectory()

        patcher = patch(
            'sohstationviewer.model.handling_data.readWaveformTrace')
        self.addCleanup(patcher.stop)
        self.mock_readWaveformTrace = patcher.start()
        self.mock_readWaveformTrace.return_value = {
            'startTmEpoch': 0,
            'endTmEpoch': 0,
        }

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_all_traces_are_processed(self):
        data_time = [12512512, 12215125121]
        readWaveformMSeed(
            self.q330_waveform_file, self.q330_waveform_file.name,
            self.station_id, self.channel_id, self.traces_info, data_time,
            self.temp_dir.name)
        self.assertEqual(len(self.traces_info), len(self.mseed_waveform_stream))

    def test_readWaveformTrace_called(self):
        data_time = [12512512, 12215125121]
        readWaveformMSeed(
            self.q330_waveform_file, self.q330_waveform_file.name,
            self.station_id, self.channel_id, self.traces_info, data_time,
            self.temp_dir.name)
        self.assertTrue(self.mock_readWaveformTrace.called)

    def test_start_data_time_earlier_than_earliest_start_time(self):
        self.mock_readWaveformTrace.return_value = {
            'startTmEpoch': 51251,
            'endTmEpoch': 2623623,
        }
        start_time = 0
        data_time = [start_time, 1625532949]

        readWaveformMSeed(
            self.q330_waveform_file, self.q330_waveform_file.name,
            self.station_id, self.channel_id, self.traces_info, data_time,
            self.temp_dir.name)
        self.assertEqual(data_time[0], start_time)

    def test_end_data_time_later_than_latest_start_time(self):
        # End time set to be the last second of 9999 so as to be later than the
        # end time of any test data.
        self.mock_readWaveformTrace.return_value = {
            'startTmEpoch': 51251,
            'endTmEpoch': 2623623,
        }
        end_time = 253402326000
        data_time = [1625443222, end_time]

        readWaveformMSeed(
            self.q330_waveform_file, self.q330_waveform_file.name,
            self.station_id, self.channel_id, self.traces_info, data_time,
            self.temp_dir.name)
        self.assertEqual(data_time[1], end_time)

    def test_data_time_is_between_earliest_start_and_latest_end_time(self):
        self.mock_readWaveformTrace.return_value = {
            'startTmEpoch': 51251,
            'endTmEpoch': 2623623,
        }
        start_time = 512579
        end_time = 2623616
        data_time = [start_time, end_time]

        expected_updated_data_time = [51251, 2623623]
        readWaveformMSeed(
            self.q330_waveform_file, self.q330_waveform_file.name,
            self.station_id, self.channel_id, self.traces_info, data_time,
            self.temp_dir.name)
        self.assertEqual(data_time, expected_updated_data_time)

    # Test that dataTime is updated correctly

